/* Convert a Protracker module into a text file and samples. 
The module should be given on stdin; the text file appears on stdout.
azz@josstix.demon.co.uk, Adam Sampson */

/* FILEMODE gives a very slight performance boost (10%). */
/* #define FILEMODE */

#include <stdio.h>
#include <stdlib.h>

unsigned short int samplelength[32];

struct note {
  char name[3];
  unsigned short int period;
} notes[] = {
  "C-1", 856, "C#1", 808, "D-1", 762, "D#1", 720, "E-1", 678, "F-1", 640,
  "F#1", 604, "G-1", 570, "G#1", 538, "A-1", 508, "A#1", 480, "B-1", 453,
  "C-2", 428, "C#2", 404, "D-2", 381, "D#2", 360, "E-2", 339, "F-2", 320,
  "F#2", 302, "G-2", 285, "G#2", 269, "A-2", 254, "A#2", 240, "B-2", 226,
  "C-3", 214, "C#3", 202, "D-3", 190, "D#3", 180, "E-3", 170, "F-3", 160,
  "F#3", 151, "G-3", 143, "G#3", 135, "A-3", 127, "A#3", 120, "B-3", 113,
  "---", 0
};
#define NUMNOTES (sizeof(notes)/sizeof(struct note))
#define ERROR(c,x) if(c){fprintf(stderr,"%s\n",x);exit(20);}

#ifdef FILEMODE
unsigned char *datain;
long datainoffset,datainmax;
#define BUFREAD(x) ERROR(datainoffset+(x)>datainmax,"Premature end of file");memcpy(buffer, datain+datainoffset, x);datainoffset=datainoffset+x;
#else
#define BUFREAD(x) fread(buffer, sizeof(unsigned char), x, stdin);ERROR(feof(stdin),"Premature end of file");
#endif

#define NBYTES(x) ((unsigned short int)(2*(buffer[(x)]*256+buffer[(x)+1])))

void main(int argc, char **argv) {
  unsigned char buffer[256],numblocks,songlength,*inst;
  int i,j,k,l,m;
  FILE *f;

#ifdef FILEMODE
  if ((f=fopen(argv[1],"r"))) {
    fseek(f,0,SEEK_END);
    datainmax = ftell(f);
    datain = malloc(datainmax+1);
    fseek(f,0,SEEK_SET);
    fread(datain,sizeof(unsigned char),datainmax,f);
    fclose(f);
  } else {
    ERROR(1,"Couldn't open input file.");
  };
#endif

  printf("Produced by mod2text v0.1\n\n");
  BUFREAD(20);
  printf("Name: %-20s\n\n", buffer);
  printf("      name                   len  f vo lpst lpln\n");
  for (i=1; i<=31; i++) {
    BUFREAD(22);
    printf("S.%02x: %-22s ", i, buffer);
    BUFREAD(8);
    samplelength[i] = NBYTES(0);
    printf("%04X %01X %02X %04X %04X\n",
	   samplelength[i], buffer[2] & 0x0F, buffer[3],
	   NBYTES(4), NBYTES(6));
  };
  printf("\n");
  BUFREAD(2);
  songlength = buffer[0];
  /* buffer[1] is always 127 */
  BUFREAD(128);
  numblocks = 0;
  for(i=0; i<songlength; i++) if (buffer[i]>numblocks) numblocks=buffer[i];
  numblocks++;
  printf("NumB: %02x\n", numblocks);
  printf("Play: ");
  for(i=0; i<songlength; i++) printf("%02x ", buffer[i]);
  printf("\n");
  BUFREAD(4);
  printf("Sign: %c%c%c%c\n\n", buffer[0], buffer[1], buffer[2], buffer[3]);

  for (i=0; i<numblocks; i++) {
    printf("\nBlock %02x\n", i);
    for (j=0; j<=63; j++) {
      printf("%02x%02x: ", i, j);
      for (k=0; k<=3; k++) {
	BUFREAD(4);
	m = 0;
	for (l=0; l<NUMNOTES; l++) {
	  if (notes[l].period == (256*(buffer[0]&0x0f)+buffer[1])) m = l;
	}; 
	/* The 4 bytes go: aaaabbbb bbbbbbbb aaaacccc cccccccc
	   where a = sample, b = period, c = effects. */
	printf("%-3s %02X %03X ", notes[m].name,
	       (buffer[0]&0xf0)+((buffer[2]&0xf0)>>4),
	       (buffer[2]&0x0f)*256+(buffer[3])
	       );
      };
      printf("\n");
    };
  };
  for(i=1; i<=31; i++) {
    if (samplelength[i]) {
      if ((f=fopen("m2t.tmp","w"))) {
	inst = malloc(samplelength[i]);
#ifdef FILEMODE
	fwrite((unsigned char *)(datain+datainoffset), sizeof(unsigned char), samplelength[i], stdout);
	datainoffset = datainoffset + samplelength[i];
#else
	fread(inst, sizeof(unsigned char), samplelength[i], stdin);
	fwrite(inst, sizeof(unsigned char), samplelength[i], f);
#endif
	free(inst);
	fclose(f);
	sprintf(buffer,"uuencode <m2t.tmp sample_data");
	printf("\nSd%02x: \n",i);
	fflush(stdout);
	system(buffer);
	system("rm -f m2t.tmp");
      } else {
	fprintf(stderr, "Unable to create temp file!\n");
	exit(20);
      };
    };
  };
#ifdef FILEMODE
  free(datain);
#endif
};
